## -*-Tcl-*- (nowrap)
 # ==========================================================================
 #  AlphaTcl - core Tcl engine
 # 
 #  FILE: "contextualMenu.tcl"
 #                                    created: 01/31/2001 {09:16:50 PM} 
 #                                last update: 03/19/2002 {10:06:16 AM} 
 #  Author: Jonathan Guyer
 #  E-mail: jguyer@his.com
 #    mail: Alpha Cabal
 #     www: http://www.his.com/jguyer/
 #  
 #  Includes contributions from Craig Barton Upright
 # 
 #  Description: 
 #  
 #  Rudimentary capabilities for Contextual Menus.  The arguments and
 #  meanings are prone to change.
 # 
 #  Copyright (c) 2001-2002 Jonathan Guyer.  All rights reserved.
 #  
 # --------------------------------------------------------------------------
 # 
 # Notes:
 # 
 # This package is designed to make it much easier for modes or other
 # packages in AlphaTcl to add/remove items from the contextual menu.  This
 # menu is completely modular -- the user can add or remove items from the
 # CM with changes taking effect immediately.  Modes, global packages, and
 # mode specific packages can all create new modules to add to the CM
 # Preferences dialog.
 # 
 # The 'context' here might be specifics about the current window, the
 # current mode, or the position and/or selection surrounding the CM click
 # point.  The items added via this package might be global or mode
 # specific, but will be added no matter what is going on in the window
 # regarding selected text and/or click position.  Items within particular
 # submenus added to the main CM, however, can certainly change depending
 # on the list of values currently stored in the variable 'alpha::CMArgs'
 # or other 'context' parameters.
 # 
 # 'Non-contextual' modules can also be added to the CM, generally global
 # submenus that might already be inserted into one of the menu bar menus
 # that users might find useful as a CM shortcut.  If these perform text
 # editing tasks within the current window, however, they must take the CM
 # click position (again, available via the list of values in
 # 'alpha::CMArgs'), and NOT on the current position.
 # 
 # The CM is listed in four sections, each separated by a divider.  The
 # first contains window specific modules defined by this package, such as
 # the current path (similar to the title bar pop-up), and the current
 # marks.  The second is for mode-specific modules.  The third includes
 # global CM modules defined by other packages in AlphaTcl.  The final
 # section includes more window specific modules that probably won't be
 # accessed very often, including mode, format, and wrap.  The CM Utils
 # module is always listed last.  The following sections explain how to add
 # modules to the second and third sections.
 # 
 # (1)  Adding Global Submenus
 # 
 # Any package can add a submenu in the third section of the CM by simply
 # defining a new contextualMenu package preference, as in:
 # 
 #   newPref f <contextualMenuSubmenuName>Menu 0 contextualMenu
 # 
 # and then declaring a 'menu build proc' using the proc: menu::buildProc
 # for a menu whose name does NOT end in Menu.  The package: notes, for
 # example, can add this:
 # 
 #   newPref f notesMenu 0 contextualMenu
 # 
 # Note that in order to add a submenu, the preference MUST end in 'Menu'.
 # 
 # This preference will then appear in the "Contextual Menu" preferences
 # page, allowing the user to add it if desired.  During the actual
 # construction of the CM, the 'Menu' string will be stripped from the name
 # of the submenu which will be added, in this example appearing as
 # 'Notes'.  In this case, the package: notes has already defined a menu
 # with this stripped name, so nothing more needs to be done here. 
 # 
 # Packages that already create submenus for Alpha's global menu bar will
 # find that adding a CM submenu involves just this one line of extra code,
 # so long as a menu build proc is registered.
 # 
 # (Only defined submenus can be added this way, not menus created for the
 # menu bar with alpha::menu or addMenu.)
 # 
 # These preferences should be declared in a package's initialization
 # script so that they're available the first time the CM is created. 
 # Because the number of potential modules is virtually unlimited, it is
 # considered bad practice to set the initial value of these preferences to
 # '1' -- simply let the user know in a help file that the module is
 # available, and allow him/her to choose to activate it.
 # 
 # If your package has not been activated the first time that the CM is
 # built (which will always be well past initialization, btw), the user's
 # module pref for your package will be unset.  If your package is not
 # likely to be globally activated but is intended for a specific mode,
 # simply declare the preference for each mode which might make use of it
 # as described below.
 # 
 # (2)  Adding Global Items
 # 
 # It is also possible to add an item that appears in the third section of
 # the main CM, although this is not recommended -- given that there will
 # be a lot of possibilities for adding CM items/submenus, the CM could get
 # very busy very quickly which kind of defeats the purpose of the whole
 # exercise.
 # 
 # Given that ...  simply define a preference WITHOUT 'Menu' at the end of
 # its name.  If the name ends with 'Item' that will be stripped from the
 # item's name before being inserted into the CM.  (The rationale here is
 # that it will be much easier for a user to know what they're adding or
 # removing in the prefs dialog if each name ends with 'Menu' or 'Item',
 # although we don't place that as a hard restriction.)
 # 
 # However, you're not done yet.  These items will not be routed through
 # a submenu's menu proc, they are with the submenus described above, but
 # rather through 'contextualMenu::contextualProc'.  In order for the
 # item's procedure to be properly called, you must define a proc with
 # either the name 
 # 
 #   <mode>::<menuItemName>
 #   
 # or
 # 
 #   contextualMenu::<menuItemName>
 # 
 # No arguments are passed -- the proc must get the CM click position via
 # the "alpha::CMArgs" variable.
 # 
 # (An alternative is to register a 'contextualMenuHandler' hook -- this is
 # passed through 'contextualMenu::contextualProc' below, along with some
 # arguments as indicated.  This has not been widely tested.)
 # 
 # (3)  Adding Mode Specific CM Menus and Items
 # 
 # Mode specific CM modules appear in the second section of the CM, and can
 # be turned on and off by the user in a separate page presented in the
 # "Contextual Menu Items" dialog.  These can be defined both by a mode or
 # by a package that might be activated by the mode.  
 # 
 # This is done just as in (1) and (2), except that the preference should
 # be declared for 'contextualMenu<mode>', as in:
 # 
 #   newPref flag someCMThing<Menu/Item> 0 contextualMenu<mode>
 # 
 # which in practice will look something like:
 # 
 #   newPref f tclUtilsMenu 0 contextualMenuTcl
 # 
 # These menus must also have a build proc declared, as in:
 # 
 #   menu::buildProc "tclUtils" Tcl::buildCMUtils
 #
 # and similar to global menus, should take the CM click position into
 # account rather than the cursor position when both building the menu and
 # executing the item.
 # 
 # While package modules should always be initially turned off, mode
 # developers have complete discretion in deciding which mode specific CM
 # modules defined in the mode's source files should be turned on by
 # default.
 # 
 # --------------------------------------------------------------------------
 # 
 # To Do:
 # 
 #  Create more mode/click position specific contextual items !!!
 # 
 # Some of the following would give developers/users more control over how
 # the items in the CM are presented, but adding more user control also
 # contributes to the idea that Alpha takes a lot of time to configure ...
 # 
 #  Implement "Help" button in prefs dialog.  Will get messy, and I'm not
 #   looking forward to it.  Perhaps creating the dialog via 'dialog::make'
 #   would make this easier, although figuring out how will take (me - cbu)
 #   a lot of time, and at the moment I'm satisfied with it as it is.
 # 
 #  Create some method to allow packages/modes to decide where in the
 #   menu the item should be placed?  The current scheme feels pretty
 #   intuitive to me -- with mode specific items being given priority over
 #   global items, and all of these lists being sorted alphabetically. 
 #   More control == longer time to build the CM ??
 #   
 #  Along with the above, perhaps a package API to declare items would be
 #   in order.  Currently, a mode/package only has to declare a 'newPref'
 #   and a 'menu build proc' to interface with the CM, which seems pretty 
 #   minimal, but perhaps a
 #   
 #     contextualMenu::register
 #   
 #   proc could be created which takes 'name' 'where' 'value' 'global/mode'
 #   arguments would be easier.  One problem with this is where to define
 #   such a proc -- we want to avoid sourcing this file until init is
 #   complete, and if a mode or package is calling the proc before this
 #   package's init script is sourced, there might be trouble.
 #   
 #  Give the user control over the order in which items are presented.
 #   More control == longer time to build the CM ??
 # 
 #  Allow global items to be turned off for specific modes.
 #   More control == longer time to build the CM ??
 # 
 # ==========================================================================
 ##

alpha::feature contextualMenu 0.7 "global-only" {
    # Initialization script.
    
    # Turn this item on to include a "Contextual Menu Utils" submenu||
    # Turn this item off to remove the "Contextual Menu Utils" submenu from
    # the contextual pop-up menu
    newPref f contextualMenuUtilsMenu   1 contextualMenu
    # Turn this item on to include the 'Format' submenu in the contextual
    # pop-up menu, which allows you to change line endings, tabsize, etc||
    # Turn this item off to remove the 'Format' submenu from the contextual
    # pop-up menu
    newPref f formatMenu                1 contextualMenu
    # Turn this item on to include the 'Mode' submenu in the contextual
    # pop-up menu, which allows you to change the mode of the current
    # window|| Turn this item off to remove the 'Mode' submenu from the
    # contextual pop-up menu
    newPref f "mode Menu"               1 contextualMenu
    # Turn this item on to include the 'Related Files' submenu in the
    # contextual pop-up menu|| Turn this item off to remove the 'Related
    # Files' submenu from the contextual pop-up menu
    newPref f relatedFilesMenu          1 contextualMenu
    # Turn this item on to include the 'Version Control System' submenu
    # in the contextual pop-up menu|| Turn this item off to remove the
    # 'Version Control System' submenu from the contextual pop-up menu
    newPref f vcsMenu                   0 contextualMenu
    # Turn this item on to include the 'Window Marks' submenu in the
    # contextual pop-up menu|| Turn this item off to remove the 'Window
    # Marks' submenu from the contextual pop-up menu
    newPref f windowMarksMenu           1 contextualMenu
    # Turn this item on to include the 'Window Path' submenu in the
    # contextual pop-up menu.  Selecting any item in the path will open a
    # 'Find File' dialog with that default location|| Turn this item off to
    # remove the 'Window Path' submenu from the contextual pop-up menu
    newPref f windowPathMenu            1 contextualMenu
    # Turn this item on to include the 'Wrap' submenu in the contextual
    # pop-up menu|| Turn this item off to remove the 'Wrap' submenu from
    # the contextual pop-up menu
    newPref f wrapMenu                  1 contextualMenu
    # Add the user interface dialog to "Config --> Packages".
    namespace eval contextualMenu {}
    menu::buildProc "contextualMenuUtils" {contextualMenu::buildCMUtils}
    proc contextualMenu::buildCMUtils {} {
	set menuList [list \
	  "Conceptual Menu Prefs" "Conceptual Menu Help"]
	return [list build $menuList {contextualMenu::cMUtils -m} {}]
    }
    menu::insert packages submenu "(-)" "contextualMenuUtils"
} {
    #Activiation script.
    hook::register   contextualMenuHook      contextualMenu::contextualMenu
} {
    # Deactivation script.
    hook::deregister contextualMenuHook      contextualMenu::contextualMenu
} uninstall {
    this-file
} maintainer {
    "Craig Barton Upright" <cupright@princeton.edu> <http://www.princeton.edu/~cupright/>
} help {
    Support for Contextual Menus, invoked by pressing 'control' and the
    mouse button simultaneously||
    Support for Contextual Menus, invoked by pressing 'control' and the
    mouse button simultaneously||
    
    Support for Contextual Menus, invoked in the MacOS by pressing 'control'
    and the mouse button simultaneously.  In the Windows OS, the CM is
    accessed by clicking the mouse's right button.
    
    Individual CM modules (menus or items) can be added or removed from the
    CM using the "Config --> Packages --> Contextual Menu Utils --> Prefs"
    menu item after this package has been activated.
    
    Default items include

	Related Files

    This submenu includes all of the files found in the directory of the
    current window.  (Similar to option titlebar clicking.)
    
	Window Path Menu

    This submenu contains a list of all folders found in the path of the
    current window.  Selecting any item will open a standard OS "Find File"
    dialog using that folder as the default.  (Similar to clicking on the
    titlebar window.)
    
	Window Marks Menu

    This submenu contains all of the items found in the Marks pop-up menu.

	VCS Menu

    This submenu contains all of the items found in the VCS pop-up menu.
    (This module is initially turned off by default.)

	Format Menu

    This submenu contains all of the items found in the 'format' pop-up
    menu in the status bar, allowing you to change tab size, line endings,
    etc.  for the current window.

	Wrap Menu

    This submenu allows you to change the current 'wrap' style of the
    current window.  See the section on wrapping in the "Alpha Manual" for
    more information.
    
	Mode Menu

    Allows you to change the mode of the current window.

	Contextual Menu Utils

    This menu contains two items, the first allowing you to change the
    Contextual Menu preferences, turning items on and off both globally and
    for mode specific items, and a "Help" menu item which opens this
    window.

	----------------------------------------------------------------
    
    In addition to these default contextual menu items, modes might have
    additional CM modules which only appear for that mode, such as a
    "Process" menu in TeX mode, or an "Html Attributes" menu which will
    list all possible options for the tag surrounding the CM click
    position.  These additional modules can be turned on and off using the
    "Config --> Packages --> Contextual Menu Utils --> Prefs" dialog, also
    available in the "Contextual Menu Utils" CM module.
    
    Some packages in AlphaTcl offer additional plug-ins to add submenus
    which might already appear in some of Alpha's global menus, such as one
    containing a list of all recent files (package: recentFilesMenu), or
    note files that you have created and saved (package: notes), etc.  The
    best way to find out what these items do is to simply activate them and
    see what is now available in the contextual menu -- items can always be
    turned off, with the changes taking effect immediately.
  
    AlphaTcl developers should read the notes in "contextualMenu.tcl" for
    more information on how to create contextual menu plug-ins.
} requirements {
    alpha::package require Alpha 8.0a2
}

# If this file is being sourced, we know that initialization is complete
# and all packages should by now have declared CM prefs.  If any CM prefs
# were saved but not created with a newPref yet, we can assume that they
# are for a package that is no longer active, so we'll unset them.  If a
# package is generally only active for a particular mode, the pref should
# be declared as
# 
# newPref flag someCMMenuOrItem 0 contextualMenu<mode>
# 
# for each possible mode for which the package is useful.

foreach cmPref [array names contextualMenumodeVars] {
    if {![lcontains allFlags $cmPref]} {
	unset contextualMenumodeVars($cmPref)
    }
}

unset cmPref

namespace eval contextualMenu {}

#  contextual menu  #

## 
 # -------------------------------------------------------------------------
 # 
 # "contextualMenu::buildContextualMenu" --
 # 
 # The main task here is to define the menu proc that will be used for the
 # top level CM items.  The actual building of the menu takes place in
 # 'contextualMenuHook'.
 # 
 # -------------------------------------------------------------------------
 ##

proc contextualMenu::buildContextualMenu {} {
    return [list build {} {contextualMenu::contextualProc} {}]
}

## 
 # -------------------------------------------------------------------------
 # 
 # "contextualMenu::contextualMenu" --
 # 
 # The main task here is add to the list of items used in creating the CM.
 # The actual building of the menu takes place in 'contextualMenuHook'.
 # 
 # -------------------------------------------------------------------------
 ##

proc contextualMenu::contextualMenu {menuName args} {

    global contextualMenumodeVars menu::additions  alpha::CMName
    
    menu::buildProc $menuName contextualMenu::buildContextualMenu
    # Add the menus/items
    foreach item [listMenuItems] {
	if {[regsub "Menu$" $item {} item]} {
	    lappend menu::additions($menuName) [list submenu end $item]
	} else {
	    regsub "Item$" $item {} item
	    lappend menu::additions($menuName) [list items end $item]
	}
    }
}

## 
 # -------------------------------------------------------------------------
 # 
 # "contextualMenu::listMenuItems" --
 # 
 # Create the list (including dividers) of CM items based on global/mode
 # prefs associated with this package.  We'll do this separately and call
 # it above just in case some other code wants to get the actual list.
 # 
 # -------------------------------------------------------------------------
 ##

proc contextualMenu::listMenuItems {} {
    
    global contextualMenumodeVars mode 
    global contextualMenu${mode}modeVars
    
    set list1      [list]
    set list2      [list]
    set list3      [list]
    set list4      [list]
    set ignoreList [list "contextualMenuUtilsMenu"]
    # This is set of items potentially appearing at the top of the cm.
    set begList [list \
      "openWindowsMenu" "relatedFilesMenu" "windowPathMenu" "windowMarksMenu"]
    eval lappend ignoreList $begList
    # This is set of items potentially appearing at the bottom of the cm.
    set endList [list \
      "formatMenu" "mode Menu" "vcsMenu" "wrapMenu"]
    eval lappend ignoreList $endList
    # This is the set of items potentially appearing in the middle of the cm.
    foreach pref [array names contextualMenumodeVars] {
	# Only add it if it doesn't appear above, below, but has been
	# declared using 'newPref' (and is activated).
	if {![lcontains ignoreList $pref] && $contextualMenumodeVars($pref)} {
	    lappend list3 $pref
	}
    }
    # Remove items as needed from the top of the list.
    foreach pref $begList {
	if {![info exists contextualMenumodeVars($pref)]} {continue}
	if {$contextualMenumodeVars($pref)} {lappend list1 $pref} 
    }
    # Add any additional mode specific items.
    if {[string length $mode]} {
	foreach pref [array name [lsort -ignore contextualMenu${mode}modeVars]] {
	    if {[set contextualMenu${mode}modeVars($pref)]} {
		lappend list2 $pref
	    } 
	}
    }
    # Remove items as needed from the bottom of the list.
    foreach pref $endList {
	if {![info exists contextualMenumodeVars($pref)]} {continue}
	if {$contextualMenumodeVars($pref)} {lappend list4 $pref} 
    }
    # Now create the final list.
    set fullList $list1
    if {[llength $list2]} {
	if {[llength $fullList]} {lappend fullList "(-)"}
	eval lappend fullList [lsort -ignore $list2]
    }
    if {[llength $list3]} {
	if {[llength $fullList]} {lappend fullList "(-)"}
	eval lappend fullList [lsort -ignore $list3]
    }
    if {[llength $list4]} {
	if {[llength $fullList]} {lappend fullList "(-)"}
	eval lappend fullList [lsort -ignore $list4]
    }
    if {$contextualMenumodeVars(contextualMenuUtilsMenu)} {
	if {[llength $fullList]} {lappend fullList "(-)"}
	lappend fullList contextualMenuUtilsMenu
    }
    return $fullList
}

## 
 # -------------------------------------------------------------------------
 # 
 # "contextualMenu::contextualProc" --
 # 
 # The proc called for all main CM items.
 # 
 # -------------------------------------------------------------------------
 ##

proc contextualMenu::contextualProc {menuName itemName} {
    
    global mode
 
    if {[info proc ::${mode}::${itemName}] != ""} {
	# Item is defined within the mode's namespace.
	::${mode}::${itemName} 
    } elseif {[info proc $itemName] != ""} {
	# Item is defined with the 'contextualMenu' namespace.
        $itemName 
    } else {
	# Hopefully item is defined via a handler hook.
	hook::callAll contextualMenuHandler $mode $menuName $itemName
    }
}

#  --------  #

#  format menu  #

menu::buildProc "format" {buildFileInfoMenu}

#  mode menu  #

# One is tempted to use buildModeMenu here, but the menu proc 'newMode'
# will inadvertantly call 'package::helpOrDescribe' if the control key is
# held down for too long.

menu::buildProc "mode " {contextualMenu::buildMode} {contextualMenu::postBuildMode}

proc contextualMenu::buildMode {} {
    return [list build [mode::listAll] {contextualMenu::mode -m} {}]
}

proc contextualMenu::postBuildMode {} {
    global mode
    markMenuItem -m "mode " $mode on
}

proc contextualMenu::mode {menuName itemName} {
    win::ChangeMode $itemName
}

#  related files menu  #

menu::buildProc "relatedFiles" {contextualMenu::buildRelated}

proc contextualMenu::buildRelated {} {
    
    if {![llength [set menuList [::OptionTitlebar]]]} {
	set menuList [list "(Current window is not a file"]
    }
    return [list build $menuList {contextualMenu::related -m}]
}

proc contextualMenu::related {menuName itemName} {
    ::OptionTitlebarSelect $itemName    
}

#  vcs menu  #

menu::buildProc "vcs" {contextualMenu::buildVcs}

proc contextualMenu::buildVcs {} {

    global menu::build_procs
    
    if {[info exists menu::build_procs(vcsMenu)] && [package::active vcs]} {
	return [eval [set menu::build_procs(vcsMenu)]]
    } else {
	return [list "(Package 'vcs' required" {"" -m}]
    }
}

#  window marks menu  #

menu::buildProc "windowMarks" {buildMarksMenu}

#  window path menu  #

# This should be the same behavior as clicking on the title bar.  It
# probably needs some revision to work properly in AlphaX.  Also works for
# remote files.

menu::buildProc "windowPath" {contextualMenu::buildPath} {contextualMenu::postBuildPath}

proc contextualMenu::buildPath {} {
    global file::separator contextualMenu::PathElements
    # In order to avoid ambiguous path elements like 
    # 
    #   iBook:alpha:test:alpha:file.txt
    # 
    # we'll go through this routine now.
    catch {unset contextualMenu::PathElements}
    if {[win::IsFile [set path [win::Current]]]} {
	set path [getTitleBarPath]
	set fs [file nativename [set file::separator]]
	set elements [lreverse [split $path $fs]]
    } else {
	set elements [list $path]
    }
    foreach item $elements {
	while {[info exists contextualMenu::PathElements($item)]} {append item " "}
	lappend menuList $item
	# Make sure that the path ends with file::separator.
	# Shouldn't be necessary if 'findFile' gets fixed.
	if {[file isdir $path]} {
	    set path [string trim $path [set file::separator]]
	    set path $path[set file::separator]
	} 
	set contextualMenu::PathElements($item) $path
	set path [file dirname $path]
    }
    return [list build "$menuList" {contextualMenu::path -c -m} {}]    
}

proc contextualMenu::postBuildPath {args} {
    
    if {![win::IsFile [set path [win::Current]]]} {
	enableMenuItem -m "windowPath" $path 0
    } 
}

proc contextualMenu::path {menuName itemName} {
    global contextualMenu::PathElements
    if {[info exists contextualMenu::PathElements($itemName)]} {
	titlebar [set contextualMenu::PathElements($itemName)]
    } else {
	status::errorMsg "Couldn't find directory for $itemName"
    }
}

#  wrap menu  #

menu::buildProc "wrap" {buildWrapMenu}

#  --------  #

#  contextual menu utils menu  #

proc contextualMenu::cMUtils {menuName itemName} {
    switch $itemName {
	"Conceptual Menu Prefs"  {global::contextualMenuItems}
	"Conceptual Menu Help"   {package::helpFile "contextualMenu"}
    }
}

namespace eval global {}

## 
 # -------------------------------------------------------------------------
 # 
 # "global::contextualMenuItems" --
 # 
 # Creates a dialog to turn on/off global contextualMenu preferences, as
 # well as those associated only with a particular mode.  Called by the
 # "Config --> Packages --> Contextual Menu Items" menu item, as well as
 # the menu item in the CM Utils pop-up menu.
 # 
 # This currently doesn't have a 'Help' button, because trying to implement
 # it hurts my brain.  Also, global items can NOT be turned off for a
 # particular mode.  We can incorporate either/both of these if there is a
 # strong user demand for such items, but otherwise it works well.
 # 
 # -------------------------------------------------------------------------
 ##

proc global::contextualMenuItems {} {
    
    global contextualMenumodeVars mode
    global contextualMenu${mode}modeVars

    set globalCMPrefs [lsort -ignore [array names contextualMenumodeVars]]
    set modeCMPrefs   [lsort -ignore [array names contextualMenu${mode}modeVars]]

    set y 30
    set d [list -T "Contextual Menu Items ..."]
    incr y 15
    set pageNames [list "Global Items" "Global Items"]
    if {[string length $mode]} {lappend pageNames "$mode Specific Items"}
    lappend d -m $pageNames 170 10 330 33
    # Add the first page
    set  y1 $y
    incr y1 5
    set col 0
    lappend box -n [lindex $pageNames 1]
    foreach pref $globalCMPrefs {
	eval lappend box [dialog::checkbox "[quote::Prettify $pref]" \
	  $contextualMenumodeVars($pref) [expr {15 + ($col * 160)}] y1]
	switch $col {
	    "2"     {set  col 0}
	    default {incr col ; incr y1 -18}
	}
    }
    incr y1 [expr {$col == 0} ? 10 : 28]
    # Add the second page
    set  y2 $y
    incr y2 5
    set col 0
    if {[string length $mode]} {
	lappend box -n [lindex $pageNames 2]
	if {![llength $modeCMPrefs]} {
	    set note "There are no mode specific contextual menu preferences to set."
	    eval lappend box [dialog::text $note 10 y2]
	} else {
	    foreach pref $modeCMPrefs {
		set Pref [quote::Prettify $pref]
		eval lappend box [dialog::checkbox [quote::Prettify $pref] \
		  [set contextualMenu${mode}modeVars($pref)] \
		  [expr {15 + ($col * 160)}] y2]
		switch $col {
		    "2"     {set  col 0}
		    default {incr col ; incr y2 -18}
		}
	    }
	}
    } 
    incr y2 [expr {$col == 0} ? 10 : 28]
    # Add the cancel/okay buttons, and evaluate the dialog.
    set y3  [expr {$y1 > $y2 ? $y1 : $y2}] ; incr y3 5
    set y4  [expr {$y3 + 20}]
    lappend d -b "OK"   420 $y3 480 $y4
    lappend d -b "Cancel" 345 $y3 410 $y4
    # Should try to implement this later.
#     lappend d -b "Help"   400 10  460 30
    incr y4 10
    if {[catch {eval dialog -w 500 -h $y4 $d $box} result]} {
	status::msg "Cancelled." ; return
    } elseif {[lindex $result 2]} {
	status::msg "Cancelled." ; return
    }
    # These are the items returned:
    # 
    #   0  the name of the most recent 'page' of the dialog.
    #   1  Okay
    #   2  Cancel
    #   
    #   and, if implemented, 
    #   
    #   3  Help
    #   
    #   which should determine what the 'count' should be intially set to.
    #   
    set count 3
    foreach pref $globalCMPrefs {
	set newValue [lindex $result $count]
	set oldValue [set contextualMenumodeVars($pref)]
	if {$newValue != $oldValue} {
	    set contextualMenumodeVars($pref) $newValue
	    prefs::modified contextualMenumodeVars($pref)
	} 
	incr count
    }
    if {[string length $mode]} {
	foreach pref $modeCMPrefs {
	    set newValue [lindex $result $count]
	    set oldValue [set contextualMenu${mode}modeVars($pref)]
	    if {$newValue != $oldValue} {
		set contextualMenu${mode}modeVars($pref) $newValue
		prefs::modified contextualMenu${mode}modeVars($pref)
	    } 
	    incr count
	}
    } 
    status::msg "Contextual Menu preferences have been saved."
}

# ===========================================================================
#
#  ------------  #
# 
#  Version History  #
# 
#  modified by  rev    reason
#  -------- --- ------ -----------
#  01-31-01 JEG 0.1    Original.
#  01-06-02 JL  0.2    Added handler of top level menu items.
#  02-10-02 cbu 0.3    Changed package name to "contextualMenu".
#                      Changed package to global-only feature in order to
#                        easily create a deactiviation script.
#                      Added more default cm submenus: "Open Windows", 
#                        "Window Path" "Window Marks"
#                      Enhanced contextualMenu::buildContextualMenu to accept
#                        items based upon contextualMenu preferences.
#                      Modifications so that any package can declare such a
#                        preference to add an item globally to the cm.
#                      Each mode now has a "<mode> Utils" menu reserved.
#  02-26-02 cbu 0.4    Added ability for 'contextualMenu::contextualProc' 
#                        to call individual items.
#                      Each mode can now have a mode specific set of preferences
#                        which can be turned on or off.  Useful for packages
#                        which might not be global but are only active for
#                        certain modes.  Also allows for easy insertion/
#                        removal of mode menu submenus.
#                      No longer using the standard package prefs dialog.
#  02-27-02 cbu 0.5    Minor dialog improvements.  
#                      Taking advantage of the fact that we're using Tcl 8.0
#                        and calling procs within namespaces if possible.
#                      Improved support code which takes 'click position'
#                        context into account to demonstrate functionality.
#                      Speed improvement by not using 'menu::insert', which
#                        causes the menu to be rebuilt each time.  Now we
#                        simply add to the CM "menu::additions" array.
#                        The menu gets built in 'contextualMenuHook'
#                      "Open Windows" module no longer part of this package,
#                        but should be defined in "openWindowsMenu.tcl"
#                      Improved documentation, which probably should be moved
#                        to "Extending Alpha" eventually.
#  03-03-02 cbu 0.6    Changed 'mode' menu name to 'mode ' to reduce conflicts
#                        with 'Mode Prefs' menu.  ("elecCompletions.tcl" adds
#                        items to the 'mode' menu which is used to created
#                        the "Mode Prefs" menu.)
#                      Support code for "Open Windows" module fixed to deal with
#                        duplicate window names, i.e. <2>, now lists them all
#                        in alphabetical order w/o key bindings.
#  03-18-02 cbu 0.7    Small fixes for Windows.
#                      Added 'Related Files' submenu, similar to option 
#                        title bar click.
#                      Added 'VCS' menu, same as on side of window bar.
#                      Modes can now define mode specific item procs within
#                        their own namespace, as in '<mode>::itemName' rather
#                        than 'contextualMenu::itemName' (which still works.)
# 

# ===========================================================================
# 
# .
